'''
descriptions.py
'''

class HasDesc: # XXX refile, rename
    """
    Base class for things with a desc (description)
    to help with initialization, copying, and saving.

    By convention, their __init__ methods should have an optional parameter
    called desc, which is a description of parameter values for the new object,
    which overrides the values it might otherwise choose for its parameters.
    They should save this in self.__desc at the start of __init__, so it can be
    used by submethods of __init__ or subsequent initialization-like methods;
    by convention, it's either None (meaning no description was given), or a
    tuple whose 0-element (desc[0]) relates to the superclass
    and should be passed to the superclass __init__ method. (So the best
    expression to pass to the superclass is (desc and desc[0]).)
    """
    # XXX some of the following method docstrings are redundant with the class docstring; they should be trimmed.
    def __init__(self,
                 desc = None ):
        "XXX bring in stuff from class docstring"
        assert desc is None or desc == ()
        
    def get_desc(self): # HasDesc
        """
        Return description data which defines our state,
        assuming our class is already known.

        By convention, when there is leaf class and superclass contribution,
        this will be a tuple whose first element concerns the superclass
        and whose other elements concern the leaf class. If there is no contribution
        from some subclass, this will have the same value in that subclass as in its
        immediate superclass (and get_desc needn't be overridden in that subclass).

        By convention, this is always a tuple, and we can add new elements
        at the end in an upward-compatible way; therefore we ignore extra elements
        at the end.

        By convention, when a class is using a desc during construction,
        it saves it in self.__desc (otherwise None) from which methods that
        need parts of it retrieve those parts.

        By convention, each subclass's __init__ method might end with a call
        to that class's __alter_from_desc method, which use self.__desc to
        modify state after the rest of __init__ concerned with that particular
        class (not a superclass or subclass). Note the double underscore in
        both the methodname __alter_from_desc and the attrname __desc.
        (This means there is no reason to define those in this base class.)
        """
        return () ###
    
    def save_if_desired(self, fileobj, indent):
        """
        If self's leaf class desires it, save self in the given fileobj
        as exactly one safe-eval expression (no identifiers -- just numbers,
         lists, tuples, and numeric dicts),
        with the given indent before each line (and \n at the end of each line),
        and return True; otherwise do nothing and return False.
        """
        # XXX NEED TO CLEAN THIS UP -- use something more general like JSON?
        classname_for_mangling = self.__class__.__name__.split('.')[-1]
        save_role_attr = '_' + classname_for_mangling + '__save_role'
        val = getattr(self, save_role_attr, None)
            # let each class define __save_role if it wants to [ ### XXX rename this attr?]
            # XXX warn if this does not uniquely determine the leaf class; note, it needn't be atomic like it is so far
        # XXX TODO: let val be a method if desired; for now it's just a constant
        if val is None:
            return False
        what = val # indicates our class and our role in container (world)   
        desc = self.get_desc()
        totaldesc = (what, desc)
        fileobj.write( indent + "%r\n" % (totaldesc,) )
        return True

    pass

