'''
files.py
'''

SAFECHARS = " ,.0123456789-[]e+" + "(){}:" # "needed here" + "safe but not yet needed here"
    # Note: this permits no identifiers (unless made entirely of 'e'), or string literals,
    # but works for lists/tuples/dicts made entirely of numbers which are printable without letters,
    # or are printable only with the letter 'e' (for printed floats). (We could dispense with the
    # latter need by better controlling the print format when we write, but there are old files
    # for which we didn't do that, and it's not too hard to make it safe, since no unsafe builtins
    # are named 'eeeee' etc.)

def safe_eval(valstring):
    # XXX could use re module to speed this up!
    # XXX or in Python 2.6, use literal_eval.
    chars = {}
    for c in valstring:
        chars[c] = True
    for c in chars:
        if not c in SAFECHARS:
            raise ValueError, "unsafe char %r" % (c,)
    # now it's safe to eval (of course it might be a syntax error);
    # in case it has identifiers made entirely of 'e's, supply an empty namespace.
    return eval(valstring, {})

class interpret_terrain_file(object):
    """
    Terrain file reader
    """

    _KEYS = "size heights colors objects".split()
    
    error = False # not yet anyway
    _error_message = ""

    # these attrs will be set with the data we read
    size = None
    heights = None
    colors = None
    objects = []

    # these are also or instead used as accessors for the data we read
    # (they're computed by our _fix_values method)
    data = None
    nx = None
    nz = None
    
    def __init__(self, filename):
        self.filename = filename
        self._interpret_file()
        
    def error_message(self):
        return self._error_message
    
    def set_error(self, msg = None):
        if not self.error: # don't overwrite the first-detected error message
            self._error_message = msg or "error"
            self.error = True
        return

    def require(self, cond, msg = None): # like assert but works with python -O
        if not cond:
            self.set_error(msg)
            raise ValueError # so caller doesn't need to return explicitly
    
    def _interpret_file(self): # needs to work in python 2.5
        try:
            fp = open(self.filename, "rU")
        except IOError:
            self.set_error("can't open file %r" % self.filename)
            return
        try:
            origlines = fp.readlines()
            # TODO: notice initial format line, verify it;
            # for current and past formats see terrain.py;
            # this reader should work up to 090507 format
            # (which is defined partly in terrain.py and partly in world.py XXX UGH @@@)
            
            lines = [] # process continuation lines, and canon them in other ways
            onelineparts = []
            
            for line in origlines:
                line = line.rstrip()
                if not line or line.startswith('#'):
                    continue
                if line.startswith(" "):
                    # continuation line
                    onelineparts.append(line)
                    continue
                # regular line
                if onelineparts:
                    lines.append("".join(onelineparts))
                    onelineparts = []
                onelineparts.append(line)
            if onelineparts:
                lines.append("".join(onelineparts))
            del onelineparts, origlines
            
            for line in lines:
                parts = line.split(' = ', 1)
                if len(parts) == 2:
                    # key/value pair
                    key, val = parts
                    if key in self._KEYS:
                        self._keyval(key, val)
                    else:
                        self.set_error("unrecognized key: %s" % key)
                        return
                else:
                    self.set_error("syntax error on line: %r" % line)
                    return
            self._fix_values()
        except:
            ## raise
            fp.close()
            self.set_error("syntax error in value, or internal bug")
        fp.close()
        return

    def _keyval(self, key, val): # doesn't need exception protection
        # same method for all keys so far:
        val = safe_eval(val)
        setattr(self, key, val)

    def _fix_values(self):
        # be careful not to rely on the assert statement for verifying the file!
        size = self.size
        heights = self.heights # check type and lengths
        colors = self.colors

        self.data = self.heights
        
        nx, nz = size
        self.nx = nx = int(nx)
        self.nz = nz = int(nz)

        require = self.require
        
        require(nx > 0 and nz > 0, "invalid size")

        # kluge
        nx += 1
        nz += 1

        n = nx * nz

        require(type(heights) == type([]), "invalid heights type")
        for row in heights:
            require(type(row) == type([]), "invalid heights row type")
            # this might be too slow:
            for d in row:
                if not (type(d) == type(0.0)):
                    require(False, "invalid heights data")
            require(len(row) == nz, "require len(row) %r == nz %r" % (len(row), nz))
        require(len(heights) == nz)

        if colors is not None: # let them be optional, since not in early files
            colors = list(colors)
            require(len(colors) == n * 4) # XXX or n * 3?
            # this might be too slow:
            for c in colors:
                if not (type(c) is type(0) and 0 <= c < 256):
                    require(False, "invalid colors data")

        # client will interpret the self.objects list
        
        return

    pass
