'''
game_handlers.py
'''

from pyglet.gl import *
import pyglet

from pyglet.window import key

from text import TextGroup

class game_handlers(object):

    fps_display = None
    label_batch = None
    
    def __init__(self, world, window, fps_counter = False):
        self.world = world
        self.window = window
        if fps_counter:
            self.fps_display = pyglet.clock.ClockDisplay()
        self.textgroup = TextGroup()

        if 0:
            # experimental text label. Not sure how many of the changes here
            # were needed to make it visible (e.g. remove on_resize, call window.clear).
            ### BUG in shadows when used, by messing up the texture clamping. 
            self.label_batch = pyglet.graphics.Batch()
            ctrl = '<ctrl>' if key.MOD_ACCEL == key.MOD_CTRL else '<cmd>'
            self.label = pyglet.text.Label('test label %s'.replace('%s', ctrl),
                                             x = 20, ## self.window.width / 2,
                                             y = 45, ## self.window.height / 2,
                                             anchor_y = 'bottom',
                                             color = (128, 128, 128, 255),
                                             batch = self.label_batch,
                                             group = self.textgroup
                                           )
    
    def on_draw(self): # XXX if this ever becomes a not-always-pushed state, factor outer draw stuff out of it
        ##glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        self.window.clear()
        
        glMatrixMode(GL_PROJECTION)
        glPushMatrix() # avoid trashing pyglet default (which is not identity in this matrix)
        glLoadIdentity()
        gluPerspective(60., self.window.width / float(self.window.height), .1, 1000.)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        self.world.draw() # warning: this modifies modelview matrix
        
        glMatrixMode(GL_PROJECTION)
        glPopMatrix() # restore pyglet default
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity() # apparently same as pyglet default

        if self.label_batch:
            self.label_batch.draw()
        if self.fps_display:
            self.fps_display.draw()
        
        # the following causes nothing to happen or be drawn, and cpu to go even higher --
        # maybe i need to change it in events too??
        ## self.window.invalid = False ### re issue 399, guess at how to put it in, evidently not needed on all macs only some

    def on_text(self, text): #bks, works
        if text == ',' or text == '<':
            self.world.camera.zoom *= 1.05**(-1)
            return pyglet.event.EVENT_HANDLED
        if text == '.' or text == '>':
            self.world.camera.zoom *= 1.05**(1)
            return pyglet.event.EVENT_HANDLED

    def on_mouse_scroll(self, x, y, dx, dy): #bks
        # this comment is so I also find this if I search for on_mouse_wheel
        self.world.camera.zoom *= 1.05**(dy) # move top of wheel same direction as camera should move (towards or away from you)

    def on_key_press(self, symbol, modifiers):

        # (note: regarding key.MOD_ALT or key.MOD_OPTION, see utilities.option_key_pressed)

        if symbol == key.F:
            self.world.skip_next_update = True
            self.window.set_fullscreen(not self.window._fullscreen)
            return pyglet.event.EVENT_HANDLED

        if symbol == key.F8:
            self.world.terrain.debug_toggle_wireframe()
            return pyglet.event.EVENT_HANDLED

        if symbol == key.J or symbol == key.SPACE:
            for p in self.world.players(): # main player, buddies, and enemies (all moving objects)
                p.control_jump() # enemies ignore this
            return pyglet.event.EVENT_HANDLED

        if symbol == key.N:
            self.world.player.control_add_buddy()
            return pyglet.event.EVENT_HANDLED

        if symbol == key.DELETE or symbol == key.BACKSPACE: # on Mac, key.BACKSPACE seems to be the DELETE key
            self.world.control_delete_buddy()
            return pyglet.event.EVENT_HANDLED

        if symbol == key.Y: # F9 is taken by Mac OS; F7 may have logged me out w/o saving changes (not sure); used to be E
            for p in self.world.players():
                p.debug_toggle_collisions()
            return pyglet.event.EVENT_HANDLED
        
        if symbol == key.E:
            self.world.player.control_add_enemy()
            return pyglet.event.EVENT_HANDLED

        if symbol == key.P:
            for p in self.world.players():
                p.pause_and_print_player_pos()
            return pyglet.event.EVENT_HANDLED

        if symbol == key.S:
            self.world.save()
            return pyglet.event.EVENT_HANDLED

        if symbol == key.C:
            self.world.toggle_colortrails()
            return pyglet.event.EVENT_HANDLED

        if symbol == key.G:
            self.world.cycle_goodguy_coloring()
            return pyglet.event.EVENT_HANDLED

        if symbol == key.TAB:
            self.world.cycle_camera()
            return pyglet.event.EVENT_HANDLED

        if 0:
            # enable this to find key symbols
            print "key press ignored:", symbol
            for x in dir(key):
                if getattr(key, x) == symbol:
                    print " (%r)" % x


