'''Game main module.

Contains the entry point used by the run_game.py script.
'''

from pyglet.gl import *
import pyglet

from pyglet.window import key

from world import World

from game_handlers import game_handlers

# ==

# some toplevel code is from pyglet/examples/graphics.py

# default startup options
WIDTH, HEIGHT = 800, 600
fullscreen = False
N = 40 # lawn size
load_filename = None # option can set this to a filename to load from
fps_counter = False # not always true, since it might be the cause of an
    # intermittent cosmetic bug (shadow-texture clamping turned off initially)

import sys
while len(sys.argv) > 1 and sys.argv[1].startswith('-'):
    if sys.argv[1] == '-f' or sys.argv[1] == '--fullscreen': # or toggle at runtime with key.F
        fullscreen = True
    elif sys.argv[1] == '--fps':
        fps_counter = True
    elif (sys.argv[1] == '-N' or sys.argv[1] == '--lawnsize') and len(sys.argv) > 2:
        N = int(sys.argv[2])
        del sys.argv[2]
    elif (sys.argv[1] == '-L' or sys.argv[1] == '--load') and len(sys.argv) > 2:
        load_filename = sys.argv[2]
        del sys.argv[2]
    else:
        print "unknown option", sys.argv[1]
    del sys.argv[1]
    continue

window_options = dict(fullscreen = fullscreen,
                      resizable = True,
                      width = WIDTH, height = HEIGHT )
if fullscreen:
    # Width and height cannot be specified with fullscreen.
    del window_options['width']
    del window_options['height']

window = None # modified below

try:
    # Try and create a window with multisampling (antialiasing)
    config = Config(sample_buffers = 1, samples = 4, 
                    depth_size = 16, double_buffer = True,)
    window = pyglet.window.Window(config = config, **window_options)
    ## print "using multisampling (antialiasing)" # this works for me
except pyglet.window.NoSuchConfigException:
    # Fall back to no multisampling for old hardware
    window = pyglet.window.Window(**window_options)
    print "note: NOT using multisampling (antialiasing)"

_keyboard = key.KeyStateHandler()
window.push_handlers(_keyboard)

class Camera(object):
    zoom = 0.5 # initial value of instance variable, modified externally
        # note: a higher value means camera is closer to target
    def __init__(self, world):
        self.world = world # warning: world.player is not constant
    def applyTransform(self):
        self.world.player.lookat(self.zoom)

def update(dt):
    world.update(dt)

pyglet.clock.schedule(update)
# I tried the following on a mac whose fan turns on from this game, but it didn't help and maybe hurt.
# no time to fully understand this or optimize it. (less than 3 hours til end of pyweek!)
## pyglet.clock.schedule_interval(update, 1/60.) #### was schedule, no time arg. check whether I'm calling the right function.
    # bks added time interval... (btw if i add one, do I then need window.invalid = False in on_draw re issue 399? no, that's separate)


def gl_setup():
    # One-time GL setup
    glClearColor(1, 1, 1, 1)
    ## glColor3f(1, 0, 0)
    glColor3f(0.0, 0.5, 0.5) # this change has no visible effect when glMaterialfv is used, or when it's not used;
        # works when lighting disabled; with glColorMaterial enabled, it affects terrain, but not (I'm pretty sure) the sphere.
        # Note that sphere has bright specularity when small but not when large (presumably due to various bugs in lighting & color).
    
    glEnable(GL_DEPTH_TEST)
    glDepthFunc(GL_LEQUAL) # necessary for overlapping shadows (as expected); no other effect is known in this game

    # Simple light setup.  On Windows GL_LIGHT0 is enabled by default,
    # but this is not the case on Linux or Mac, so remember to always 
    # include it.

##    glEnable(GL_LIGHTING)
    # if I disable lighting here, then I see my colored sphere, and glColor3f above (flat) affects terrain.
    # (general OpenGL lighting/color question: how do I get the effect of *both* lighting and per-vertex coloring?
    #  (while using Batch, not glMaterialfv)

    # hmm, maybe this tells me how: http://www.sjbaker.org/steve/omniv/opengl_lighting.html
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE)
        # this says glColor or vertex colors set (these two kinds of) material colors for lighting
    glEnable(GL_COLOR_MATERIAL)
    glEnable(GL_LIGHTING)

    
    glEnable(GL_LIGHT0)

    # Define a simple function to create ctypes arrays of floats:
    def vec(*args):
        return (GLfloat * len(args))(*args)

    glLightfv(GL_LIGHT0, GL_POSITION, vec(.5, .5, 1, 0))
    glLightfv(GL_LIGHT0, GL_DIFFUSE, vec(1, 1, 1, 1))

# ==

world = None

def main():

    gl_setup()

    global world
    
    batch = pyglet.graphics.Batch()
    world = World(N, batch = batch, load_filename = load_filename)
    world.keyboard = _keyboard # for controlling the players (see world & player update methods)
    world.camera = Camera(world)

    window.push_handlers(game_handlers(world, window, fps_counter = fps_counter))
    
    pyglet.app.run()
