from pyglet.gl import *
import pyglet
from pyglet.window import key

from constants import * # kluge; some are overridden below

from textures import USE_SHADOW_TEXTURE, shadow_texture

## this won't look good unless we put normals in Triangle3.draw_SLOW: SHADOW_COLOR = (0.1, 0.4, 0.1, 0.5)
SHADOW_COLOR = (0, 0, 0, 0.5) # works alone or with shadow_texture

from terrain import Terrain

from player import Player, Buddy, Enemy
from text import Text

from euclid import Point3, Plane, Vector3
ORIGIN = Point3(0,0,0) # also in constants.py with float coords
VZERO = Vector3(0.,0.,0.)
UP = Vector3(0.,1.,0.) # also in constants.py 

from math import pi, sin, cos

from files import interpret_terrain_file


def clamp(x, x1, x2): # warning: differs from the other def of this
    # note: if range is inconsistent, x1 (minumum allowed)
    # takes precedence over x2 (max allowed)
    return max(x1, min(x2, x))

class PlayerGroup(pyglet.graphics.Group):
    "a single group to hold all players"
    def set_state(self):
        glEnable(GL_CULL_FACE) # (this doesn't seem to make any difference)
        if TRANSPARENT_PLAYERS:
            glEnable(GL_BLEND) # BUG, mystery: this is not working unless we
                # get rid of glPushAttrib in the other place using GL_BLEND,
                # but it does work for buddy balls!
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
            ## glColor4f(1,1,1,1) # this doesn't seem to have any effect
    def unset_state(self):
        glDisable(GL_CULL_FACE)
        if TRANSPARENT_PLAYERS:
            glDisable(GL_BLEND)
    pass


def draw_player_shadow_SLOW(player, terrain): # XXX move into a terrain method, in generalized form # XXX is speed an issue??

    # parameters of the 2d circular shadow:
    center = player.pos
    radius = player.collision_radius # XXX player.graphics_radius?

    # set GL state
    RAISE = 0.0
        # Note: RAISE = 0.01 also works fine and is very hard to notice as a gap;
        # 0.0 didn't work at first, but it does now since we revised glDepthFunc to GL_LEQUAL,
        # so we'll try removing this RAISE and see if there's ever an issue.
        # (Note that we need depth testing even with alpha enabled, for shadows behind hills.)
    glTranslatef(0,RAISE,0) # make sure its visible (kluge)

    glPushAttrib(GL_ENABLE_BIT)
    glEnable(GL_BLEND)
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA) # (required -- enable alone not enough)
    glColor4f(*SHADOW_COLOR)
        # Q. will this interact properly with texture? (multiply their alphas, blend their color?)
        # A. evidently it does! I guess this relates to the texture combining mode specifiable with glTexEnv.
        # Q. Would it be more efficient to use GL_DECAL (not default GL_MODULATE (??)) and avoid turning on blending?
        #    For more info see red book p. 439 and 413 (and maybe other places)...
        # A. Guess: probably not, and not sure it could even be used for that -- decal would be useful if
        #    I wanted to draw the base lawn color and the shadow at the same time.

    if USE_SHADOW_TEXTURE:
        tex = shadow_texture
        glEnable(tex.target)
        # Q: GL_CLAMP_TO_EDGE vs. GL_CLAMP? No difference seen, so I'll use the more common on. XXX find out doc difference...
        glTexParameteri(tex.target, GL_TEXTURE_WRAP_S, GL_CLAMP)
        glTexParameteri(tex.target, GL_TEXTURE_WRAP_T, GL_CLAMP)
        # XXX also set other texture modes? maybe don't need filtering, mipmaps, since it looks ok.
        # minor bug (not important): shadow intensities add (rather than max)
        glBindTexture(tex.target, tex.id)
        
        # multiplier to change radius to 114/128 * 0.5 (based on circle size in "shadow-114-128.png"):
        texscale = 1./radius * 114./128 * 0.5
        def tex_coord_func(point, texscale = texscale):
                # XXX could use glTexGen with GL_OBJECT_LINEAR instead, for GL_S and GL_T;
                # also need to glEnable(GL_TEXTURE_GEN_S) and _T. Red book pp 422-428.
            "turn point into u,v (floats from 0 to 1)"
            # We use a texture of a black-with-alpha circle on a transparent background.
            # The circle radius is fixed relative to the texture size, of course.
            # I will clamp the tex coords at the edges, so no need to worry about needing
            # the texture to be much bigger than the circle.
            offcenter = player.pos - point
            x, y_, z = offcenter
            u = 0.5 + x * texscale
            v = 0.5 + z * texscale
            return u,v
    else:
        tex_coord_func = None

    _old_vslope = terrain.VIRTUAL_SLOPE # kluge
    terrain.VIRTUAL_SLOPE = player.terrain_virtual_slope
        
    terrain.draw_subset_SLOW(
        center, radius,
        tex_coord_func = tex_coord_func
     )
    terrain.VIRTUAL_SLOPE = _old_vslope
    
    # unset state
    if USE_SHADOW_TEXTURE:
        glDisable(tex.target)
    # glDisable(GL_BLEND) messes up TRANSPARENT_PLAYERS, so do glPopAttrib instead (@@@ bug: that doesn't yet fix it):
    ## glDisable(GL_BLEND) # XXX assumes we don't use it elsewhere, or normally keep it off
    glPopAttrib()
    glTranslatef(0,-RAISE,0)

class Options:
    "owned by World, visible to players"
    # XXX TODO: put more state in here, including state that's now per-player but shouldn't be
    enable_colortrails = False # or True
    goodguy_coloring = None # or 'dirt' or 'grass'


class World:
    keyboard = None # should be set externally (kluge) to a KeyStateHandler
    camera = None # both set and used externally (not sure if kluge or ok)
    _players = _enemies = ()
    skip_next_update = False

    def __init__(self, N, batch, group = None, load_filename = None):
        self.options = Options()

        copy_from = None
        if load_filename:
            obj = interpret_terrain_file(load_filename)
            if obj.error:
                print obj.error_message()
                import sys
                sys.exit(1)
            else:
                copy_from = obj
        
        self.terrain = Terrain( ORIGIN, Point3(1,0,0), Point3(0,0,1), N, N,
                                noise = 0.1,
                                copy_from = copy_from )
        self.terrain.add_to_batch(batch, group)

        self.player = Player()
        self.player.add_to_batch(batch, group)
        self._players = [self.player]
        self._enemies = []
        self._props = []
        self.batch = batch
        self.group = group
        self.playergroup = PlayerGroup(self.group)

        if copy_from:
            self._unrecognized_object_types = {}
            for desc in copy_from.objects:
                self.add_object_described_by(desc)
            if self._unrecognized_object_types:
                print "fyi: unrecognized object types:", self._unrecognized_object_types.keys() # XXX sorted
            self._unrecognized_object_types = {}

##        if 1:
##            # just to see if it works (controls might affect both of them for now!
##            # some do, some don't, depending on implem -- e.g. motion does, jump doesn't, key.Y toggle doesn't.)
##            # (and the players have no interaction for now)
##            self.player2 = Player(Point3(20,6,20))
##            self.player2.add_to_batch(batch, group)

    def save(self): # experimental, unfinished
        import tempfile, os
        fd, filename = tempfile.mkstemp(suffix = '.outlawn', prefix = 'terrain-', text = True)
        fileobj = os.fdopen(fd, 'w')
            # XXX I hope this is ok; with no mode argument it made fileobj.write say IOError: [Errno 9] Bad file descriptor;
            # with mode 'w' it's working, at least on this Mac. It'd be nice to find docs that give the precise legality
            # conditions. Note that doc for tempfile.mkstemp says "value of text might not matter".
            # Note: the actual location of the file differs on two Macs which are similar in many ways.
        self.terrain.save_in_file_object(fileobj)
        fileobj.write("objects = [\n")
        for player in self.players():
            # XXX @@@ this really needs simplification...
            did = player.save_if_desired(fileobj, " ")
            if did:
                fileobj.write(" ,\n")
        fileobj.write(" ]\n")
        fileobj.write("# end\n")
        fileobj.close()
        print "saved terrain in %r; move it elswhere if you want to keep it!" % filename
        # TODO: some visible confirmation inside the game, e.g. temporary text at bottom of screen

    def draw(self):
        self.camera.applyTransform()
        self.batch.draw()
        self.draw_outside_of_batch()
        
    def add_object_described_by(self, desc):
        what, desc = desc #### XXX need exception protection
        if what == 1: # enemy ball, 090508 desc format
            self._add_enemy(desc = desc)
        elif what == 2: # text experiment, 090509 desc format
            self._add_text(desc = desc)
        else:
            self._unrecognized_object_types[what] = what
        
    def toggle_colortrails(self):
        self.options.enable_colortrails = not self.options.enable_colortrails
        self.options.goodguy_coloring = self.options.enable_colortrails and 'dirt' or None
        
    def cycle_goodguy_coloring(self):
        if self.options.goodguy_coloring == 'grass':
            self.options.goodguy_coloring = None
        else: # None or 'dirt'
            self.options.goodguy_coloring = 'grass'
    
    def players(self):
        return list(self._players) + list(self._enemies)

    def cycle_camera(self):
        if len(self._players) > 1:
            last_player = self._players[-1]
            self._players = [last_player] + self._players[:-1]
            self.player = self._players[0]
        
    def draw_outside_of_batch(self):
        for p in self.players():
            draw_player_shadow_SLOW(p, self.terrain)

    def control_delete_buddy(self):
        if len(self._players) > 1:
            old = self._players[-1]
            del self._players[-1]
            old.remove_from_batch()
        
    def inter_player_force_pair(self, p1, p2):
        ## return VZERO, VZERO + UP * GRAVITY * 0.9 # testing stub
        # note: kluge: if it helps, you can assume p1 comes before p2 in self.players().
        # we don't use this yet except in the testing stub commented out above.
        force_onset_distance_2 = (p1.collision_radius + p2.collision_radius) ** 2
        p1_to_p2_vec = p2.pos - p1.pos
        actual_distance_2 = p1_to_p2_vec.magnitude_squared()
        if actual_distance_2 < force_onset_distance_2:
            pseudodepth = (force_onset_distance_2 - actual_distance_2)
            if pseudodepth > 0.5:
                pseudodepth = 0.5
            force = (pseudodepth/0.2) ** 3 * GRAVITY # pushing force between players
            p1_to_p2_direction = p1_to_p2_vec.normalized()
            f2 = p1_to_p2_direction * force # p1 pushes on p2
            f1 = - f2 # and vice versa
            return f1, f2
        else:
            return VZERO, VZERO
    
    def update(self, dt):
        if self.skip_next_update:
            # time-consuming external operations should set this to avoid
            # artificially high dt values causing errors
            # (might not be needed if MAX_SAFE_TIMESTEP works well enough)
            self.skip_next_update = False
            return

        if dt > MAX_SAFE_TIMESTEP:
            if dt > 1.0:
                print "limiting large dt %r to %r" % (dt, MAX_SAFE_TIMESTEP)
            dt = MAX_SAFE_TIMESTEP # better slow-motion than terrible physics errors
            # note: this happens if I sleep and wake the computer!
            ## limiting large dt 25.684921979904175 to 1.0:
            ## bug: player fell out at Point3(16.49, -20.99, 20.78) (restoring initial position)
            # trying to play at various constant dts, even 0.15 is unsafe -- never stabilizes,
            # fallouts occur! 0.1 is safe, so setting MAX_SAFE_TIMESTEP to that for now.

        if dt == 0.0:
            ## print "dt is 0!"
            # this happens once when we start;
            # intercept it to prevent division by 0
            return
        
        # first compute inter-player force pairs (usually equal and opposite)
        # (see comment in player verlet update method for why we do this separately)

        force_from_others = {} # player -> force on it from other players
        if not self.player.paused: # kluge
            interplayer_force_pairs = {}
                # maps (p1, p2) -> (f1, f2), where f1 = force on p1 from p2;
            def store(p1, p2, f1, f2):
                # we store two entries at once, to prevent redundant computations
                interplayer_force_pairs[p1, p2] = (f1, f2)
                interplayer_force_pairs[p2, p1] = (f2, f1)

            for p1 in self.players():
                for p2 in self.players():
                    if p2 is not p1 and not interplayer_force_pairs.has_key((p1, p2)):
                        f1, f2 = self.inter_player_force_pair(p1, p2)
                        store(p1, p2, f1, f2)

            for p1 in self.players():
                f1tot = Vector3(0.,0.,0.) # not VZERO, since we modify it
                for p2 in self.players():
                    if p2 is not p1:
                        f1, f2 = interplayer_force_pairs[p1, p2]
                        f1tot += f1
                force_from_others[p1] = f1tot
        else:
            for p in self.players():
                force_from_others[p] = VZERO

        # now update player states & positions, and perform player side effects on lawn

        for player in self.players():

            what_to_do = player.update(dt,
                                       self.options,
                                       self.terrain,
                                       self.keyboard,
                                       force_from_others = force_from_others[player] )
            if what_to_do is None:
                what_to_do = {}

            make_hill = what_to_do.get('make_hill', False)
            make_depression = what_to_do.get('make_depression', False)
            stronger = what_to_do.get('stronger', False) # affects some commands:
                # wider color trails, deeper holes, higher hills
            recolor = what_to_do.get('recolor', False) # false or new color
            add_buddy = what_to_do.get('add_buddy', False) # false or its initial position
            add_enemy = what_to_do.get('add_enemy', False) # false or its initial position

            if what_to_do.get('I_fell_out', False):
                # note: intent is that only enemies can fall out, but physics bugs do it
                # to others too, when timestep is long (e.g. create enough buddies for 1 sec
                # update), so we tolerate that and handle it specially.
                # XXX should be more object-oriented (nontrivial, since players don't know
                # which of our list attrs they're in)
                ## player.fell_out(self)
                remove = False
                if player is self.player:
                    # special case for main player (bug)
                    print "bug: player fell out at %r (restoring initial position)" % player.pos
                    if len(self._players) > 2:
                        print "suggest DELETE key to remove extra buddies"
                    player.__init__() # serious kluge, but looks correct so far; seems to work
                elif player in self._enemies:
                    # this is routine
                    self._enemies.remove(player)
                    remove = True
                elif player in self._players:
                    # buddy; bug
                    print "bug: buddy fell out at %r" % player.pos
                    if len(self._players) > 2:
                        print "suggest DELETE key to remove extra buddies"
                    self._players.remove(player)
                    remove = True
                if remove:
                    player.remove_from_batch()
            
            if recolor:
                # for debugging, or for fun, recolor the under-player quad
                radius = stronger and 1 or 0 # XXX using stronger here is mainly an experiment; not sure it's a good effect
                    # radius = 0: only one quad
                self.terrain.recolor_vertices_near(player.pos, radius, recolor) 
            
            if make_hill:
                def heightfunc(old, dx, dz): # note: uses old only for clamp limits
                    goal = max(0., (9. - (dx*dx+dz*dz))) * 0.3
                    rate = dt * 2. * goal/3.
                    if stronger:
                        goal *= 2.
                        # rate is unchanged, tho relative to goal it's now lower
                    return clamp( old + rate, old, goal )
                self.terrain.raise_vertices_near( player.pos, 5., heightfunc)

            if make_depression:
                # bugs: not symmetric with make_hill (maybe the 0.?); doesn't undo it either (need rate proportional to goal?)
                def heightfunc(old, dx, dz):
                    goal = max(0., (9. - (dx*dx+dz*dz))) * 0.3
                    rate = dt * 2. * goal/3.
                    if stronger:
                        goal *= 2.
                    return - clamp( - old + rate, - old, goal )
                self.terrain.raise_vertices_near( player.pos, 5., heightfunc)

            if add_buddy:
                new = Buddy(add_buddy, paused = self.player.paused, heading = player.heading)
                batch = self.batch
                group = self.playergroup
                new.add_to_batch(batch, group)
                self._players.append(new)

            if add_enemy: ### change how this is done - controlled by level, when you change levels; for now, diff key
                initpos = add_enemy
                new = Enemy(initpos, paused = self.player.paused)
                self._insert_enemy(new)
                
            pass
        
        return

    def _add_enemy(self, desc):
        "add an enemy which fits the given description"
        new = Enemy(desc = desc, paused = self.player.paused)
        self._insert_enemy(new)

    def _insert_enemy(self, new):        
        new.add_to_batch(self.batch, self.playergroup)
        self._enemies.append(new)

    def _add_text(self, desc):
        new = Text(desc) #### UNFINISHED
        new.add_to_batch(self.batch, self.group) # XXX self.textgroup ?
        self._props.append(new)
    
    pass
